try:
    from functools import cmp_to_key
except ImportError:
    def cmp_to_key(mycmp):
        """Convert a cmp= function into a key= function"""

        class K(object):
            def __init__(self, obj, *args):
                self.obj = obj

            def __lt__(self, other):
                return mycmp(self.obj, other.obj) < 0

            def __gt__(self, other):
                return mycmp(self.obj, other.obj) > 0

            def __eq__(self, other):
                return mycmp(self.obj, other.obj) == 0

            def __le__(self, other):
                return mycmp(self.obj, other.obj) <= 0

            def __ge__(self, other):
                return mycmp(self.obj, other.obj) >= 0

            def __ne__(self, other):
                return mycmp(self.obj, other.obj) != 0

        return K

from ordereddict import OrderedDict
import semver

# Update this table and the affected unit tests when new
# parts and/or core versions are added.
#
# Note that we're using an ordered dict to maintain the (roughly chronological) presentation
# order of the modules (parts) as they were originally presented by SNAPbuild.
COMPATIBILITY_TABLE = OrderedDict([
    ("RF100",       ("2.4.19", "2.5.6")),
    ("RF200PF1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF200PD1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("SM200PU1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("SM200P81",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF200PU1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF200P81",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF266PC1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF266PU1",    ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("SS200",       ("2.4.34", "2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("RF220SU",     ("2.7.1", "2.7.2", "2.8.1")),
    ("RF220SU-EU",  ("2.8.1",)),
    ("SM220UF1",    ("2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("SM220UF1-EU", ("2.8.1",)),
    ("RF220UF1",    ("2.5.6", "2.6.2", "2.6.9", "2.7.1", "2.7.2", "2.8.1")),
    ("SN220",       ("2.7.1", "2.7.2", "2.8.1"))
])

_cores_to_parts = None
"""Maps core versions to supported parts"""


def _get_cores_to_parts():
    """Lazily builds this mapping"""
    global _cores_to_parts
    if _cores_to_parts is None:
        _cores_to_parts = {}
        for part, cores in COMPATIBILITY_TABLE.iteritems():
            for core in cores:
                if core not in _cores_to_parts:
                    _cores_to_parts[core] = set()
                _cores_to_parts[core].add(part)
    return _cores_to_parts


class UnknownCoreError(Exception):
    """Raised when presented with an unknown SNAP core version.

    :param str core: the unrecognized SNAP core version
    """

    def __init__(self, core):
        self.core = core
        Exception.__init__(self, "Unknown SNAP core version: {core}".format(core=core))


class UnknownPartError(Exception):
    """Raised when presented with an unknown SNAP module part name.

    :param str part: the unrecognized SNAP module part name
    """

    def __init__(self, part):
        self.part = part
        Exception.__init__(self, "Unknown SNAP module part name: {part}".format(part=part))


def get_cores():
    """Return the available SNAP core versions.

    Example::

        >>> snap_modules.get_cores()
        ('2.5.6', '2.4.19', '2.6.2', '2.6.9', '2.4.34', '2.7.1', '2.7.2')

    :rtype: tuple(str)
    :returns: a tuple of the available SNAP core versions
    """
    return tuple(sorted(_get_cores_to_parts().keys(), key=cmp_to_key(semver.compare)))


def get_parts():
    """Return the available SNAP module part names.

    Example::

        >>> snap_modules.get_parts()
        ('RF266PU1', 'RF200PF1', 'RF220UF1', 'RF200P81', 'SS200', 'SM220UF1', 'SM200P81', 'RF100', 'SN220', 'SM200PU1', 'RF200PU1', 'RF266PC1', 'RF220SU', 'RF200PD1')

    :rtype: tuple(str)
    :returns: an tuple of the available SNAP module part names
    """
    return tuple(COMPATIBILITY_TABLE.keys())


def get_supported_cores_for_part(part):
    """Given a SNAP module part name, return the supported SNAP core versions.

    Example::

        >>> snap_modules.get_supported_cores_for_part("SM200PU1")
        ('2.4.34', '2.5.6', '2.6.2', '2.6.9', '2.7.1', '2.7.2')

    :param str part: a valid SNAP module part name, such as "RF200P81"
    :rtype: tuple(str)
    :returns: a tuple of the supported SNAP core versions for this part name
    :raises UnknownPartError: if this part is unknown
    """
    if part in COMPATIBILITY_TABLE:
        return tuple(sorted(COMPATIBILITY_TABLE[part], key=cmp_to_key(semver.compare)))
    raise UnknownPartError(part)


def get_supported_parts_for_core(core):
    """Given a SNAP core version number, return the SNAP module part names that support it.

    Example::

        >>> snap_modules.get_supported_parts_for_core("2.7.2")
        ('RF266PU1', 'RF200PF1', 'RF220UF1', 'RF200P81', 'SS200', 'SM220UF1', 'SM200P81', 'RF200PU1', 'SN220', 'RF266PC1', 'RF220SU', 'SM200PU1', 'RF200PD1')

    :param str core: a valid SNAP core version number, such as "2.7.2"
    :rtype: tuple(str)
    :returns: a tuple of the part names that support this core version
    :raises UnknownCoreError: if this core version is unknown
    """
    parts_for_core = _get_cores_to_parts()
    if core in parts_for_core:
        return tuple(parts_for_core[core])
    raise UnknownCoreError(core)


def is_supported_combination(part, core):
    """Return True if this pair is a supported combination.

    Given a SNAP module part name and a SNAP core version number,
    return True if the specified core version is supported for the
    specified part.

    Example::

        >>> snap_modules.is_supported_combination(part="RF100", core="2.5.6")
        True
        >>> snap_modules.is_supported_combination(part="RF100", core="2.6.2")
        False

    :param str part: a valid SNAP module part name, such as "RF200P81"
    :param str core: a valid core version number, such as "2.7.2"
    :rtype: boolean
    :returns: True if the pair is a supported combination, otherwise False
    :raises UnknownPartError: if the part name is unknown
    :raises UnknownCoreError: if the core version is unknown
    """
    if core not in get_cores():
        raise UnknownCoreError(core)
    if part in COMPATIBILITY_TABLE:
        return core in COMPATIBILITY_TABLE[part]
    raise UnknownPartError(part)
